#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#define F_CPU 16000000UL
#include <util/delay.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* Debug logs for development purpose */
//#define DEBUG 1
#ifdef DEBUG
	#define DevelDebug(arg...)  printf("[DEBUG] "arg)
#else
	#define DevelDebug(arg...)
#endif

/* Character in buffer */
#define TRX_FRAME_BUFFER(index) (*(volatile uint8_t *)(0x180 + (index)))
/* Address in buffer */
#define TRX_BUFFER_ADDR(index) ((volatile uint8_t *)(0x180 + (index)))

/* The source(node address) of the received packet */
volatile uint8_t rxSender;
/* The destination(node address) of the received packet */
volatile uint8_t rxRecipient;
/* The length of the received packet */
volatile uint8_t rxLen;
/* Time in ms when the packet was received */
volatile uint32_t rxTime;

uint8_t nod1_address __attribute__((section(".data"))) = 0;
uint8_t node_address __attribute__((section(".data"))) = 0;
uint8_t nod2_address __attribute__((section(".data"))) = 0;

uint32_t symbol_threshold = 0x00000000;

void usart_0_write_bw(unsigned char c, FILE *stream);
FILE uart_output = FDEV_SETUP_STREAM(usart_0_write_bw, NULL, _FDEV_SETUP_WRITE);

#define MAX(a, b) ((a) > (b) ? (a) : (b))

// When enabled, only wakeup times will be synced between nodes
// When disabled, nodes will sync their clock without changing
// sleep times.
//#define WAKEUP_SYNC 1

#define MIN_SLEEP_MS  2000

#define FLAG_NONE 0
#define FLAG_RTS  1
#define FLAG_CTS  2
#define FLAG_TIME 3

#define TTL_MAX  10

typedef struct frame {
    uint8_t src;
    uint8_t dst;
    uint8_t flags;
    uint8_t ttl;
    char data[0];
} frame_t;

typedef struct route {
    /* Destination node */
    uint8_t dst;
    /* Next hop */
    uint8_t next;
    /* Hop count */
    uint8_t hops;
    /* Local time in ms when the last TIME packet was received */
    uint32_t this_time;
    /* Remote time in ms for "dst" */
    uint32_t dst_time;
    /* Sleep time for "dst" */
    uint16_t dst_sleep;
    /* 0 means not used */
    uint8_t used;
} route_t;

/* Length of next sleep (ms) */
uint32_t next_sleep;

/* Frame mapped over the received buffer */
volatile frame_t *rxFrame;

#define MAX_ROUTES 5
route_t      rtable[MAX_ROUTES];

uint8_t find_route(uint8_t dst);
void updateNeighbourTable(uint8_t sender, frame_t *f);

ISR(SCNT_CMP2_vect)
{
}

void initializeSymbolCounter()
{
    ASSR |= _BV(AS2); // enable asynchronous mode, with external oscillator (32.768kHz in our case)

    SCOCR1HH = (symbol_threshold >> 24);
    SCOCR1HL = (symbol_threshold & 0x00ff0000) >> 16;
    SCOCR1LH = (symbol_threshold & 0x0000ff00) >>  8;
    SCOCR1LL = (symbol_threshold & 0x000000ff);
    SCCR0 = _BV(SCEN) ;//| _BV(SCCKSEL); 	// enable symbol counter, with TOSC1/2 clock (32.768kHz)
    SCCNTHH = 0x00;
    SCCNTHL = 0x00;
    SCCNTLH = 0x00;
    SCCNTLL = 0x00;

    while (SCSR & _BV(SCBSY));
    SCIRQM = _BV(IRQMCP1); 			// enable compare match 1 IRQ*/
}

uint32_t get_symbol_counter()
{
    return (uint32_t)SCCNTLL | ((uint32_t)SCCNTLH << 8L) |
           ((uint32_t)SCCNTHL << 16L) | ((uint32_t)SCCNTHH << 24L);
}


void set_symbol_counter(uint32_t value)
{
    SCCNTHH = (value >> 24) & 0xFF;
    SCCNTHL = (value >> 16) & 0xFF;
    SCCNTLH = (value >> 8) & 0xFF;
    SCCNTLL = value & 0xFF;
}

/* Get current time in ms */
uint32_t get_time_ms()
{
    return get_symbol_counter() / 62.5;
}

void delaySymbolCounter(uint8_t seconds)
{
    symbol_threshold += (((uint32_t)seconds) * 62500);

	SCOCR1HH = (symbol_threshold >> 24);
    SCOCR1HL = (symbol_threshold & 0x00ff0000) >> 16;
    SCOCR1LH = (symbol_threshold & 0x0000ff00) >>  8;
    SCOCR1LL = (symbol_threshold & 0x000000ff);

    while (SCSR & _BV(SCBSY));
}

void delaySymbolCounterMS(uint16_t ms)
{
    symbol_threshold += (((uint32_t)ms) * 62500 / 1000);

	SCOCR1HH = (symbol_threshold >> 24);
    SCOCR1HL = (symbol_threshold & 0x00ff0000) >> 16;
    SCOCR1LH = (symbol_threshold & 0x0000ff00) >>  8;
    SCOCR1LL = (symbol_threshold & 0x000000ff);

    while (SCSR & _BV(SCBSY));
}

void initUart(void)
{
    // initializati USART 0, cu intrerupere la terminarea receptiei
    UCSR0B = _BV(RXEN0) | _BV(TXEN0);

    // parametrii sunt baud rate dat de UART_BAUD_SELECT, 0 biti de
    // paritate, un bit de stop si 8 biti de date
    UCSR0C = _BV(UCSZ01) | _BV(UCSZ00);
    UBRR0L = 103;
    UBRR0H = 0;

    stdout = &uart_output;
}

ISR(USART0_RX_vect)
{
    //returnati caracterul primit
}

void usart_0_write_bw(unsigned char c, FILE *stream)
{
    if (c == '\n') {
        usart_0_write_bw('\r', stream);
    }
    /* Wait for empty transmit buffer */
    while ( !( UCSR0A & (1<<UDRE0)) );
    /* Put data into buffer, sends the data */
    UDR0 = c;
}

void printString(char * s)
{
    int i, n = strlen(s);

    for (i=0;i<n;i++)
        usart_0_write_bw(s[i], stdout);
}

void printStringLn(char * s)
{
    int i, n = strlen(s);

    for (i=0;i<n;i++)
        usart_0_write_bw(s[i], stdout);
    usart_0_write_bw('\n', stdout);
}

void initLeds()
{
    DDRB = 0xff;
    PORTB = 0x00;
}

void setState(uint8_t state)
{
    TRX_STATE &= 0xe0;
    TRX_STATE |= state;

    while ((TRX_STATUS & STATE_TRANSITION_IN_PROGRESS) == STATE_TRANSITION_IN_PROGRESS);
}

/* Send a short frame with a magic payload */
void sendFrame(int value)
{
    setState(CMD_PLL_ON);

    TRX_FRAME_BUFFER(0) = 4;	//length - minimum length is 3
    TRX_FRAME_BUFFER(1) = 1;
    TRX_FRAME_BUFFER(2) = value;

    // start transmission
    TRX_STATE = CMD_TX_START;
    while ((TRX_STATUS & STATE_TRANSITION_IN_PROGRESS) == STATE_TRANSITION_IN_PROGRESS);
}

/* Upon receipt make a simple verification of correctness */
ISR(TRX24_RX_END_vect)
{
    rxSender = TRX_FRAME_BUFFER(0);
    rxRecipient = TRX_FRAME_BUFFER(1);
    rxLen = TST_RX_LENGTH;
    PORTB ^= (1<<PB5);

    rxFrame = (frame_t*)TRX_BUFFER_ADDR(2);
    rxTime = get_time_ms();
}

ISR(TRX24_CCA_ED_DONE_vect)
{
    rxSender = TRX_FRAME_BUFFER(0);
    rxRecipient = TRX_FRAME_BUFFER(1);
    rxLen = TST_RX_LENGTH;
    PORTB ^= (1<<PB5);

    rxFrame = (frame_t*)TRX_BUFFER_ADDR(2);
    rxTime = get_time_ms();
}

void rfInit(void)
{
    setState(CMD_TRX_OFF);

    IRQ_STATUS = 0xff;
    IRQ_MASK_struct.rx_end_en = 1;
    sei();

    //PHY_CC_CCA_struct.channel = 0x0b;
    TRX_CTRL_2_struct.rx_safe_mode = 1;
}

#define STATE_TX 1
#define STATE_RX 2

void beginSend()
{
    setState(CMD_PLL_ON);
}

void finishSend()
{
    TRX_STATE = CMD_TX_START;
    while ((TRX_STATUS & STATE_TRANSITION_IN_PROGRESS) == STATE_TRANSITION_IN_PROGRESS);
}

void sendBroadcast(uint8_t src, uint8_t dst, uint8_t ttl)
{
    frame_t *f;

    beginSend();

    f = (frame_t*)TRX_BUFFER_ADDR(3);
    /* frame size + 2 for CRC + 2 for link level address */
    TRX_FRAME_BUFFER(0) = sizeof(*f) + 2 + 2;
    TRX_FRAME_BUFFER(1) = node_address;
    TRX_FRAME_BUFFER(2) = 0xFF;

    f->src = src;
    f->dst = dst;
    f->ttl = ttl;

    finishSend();
}

/* Send a frame from @src to @dst using the next hop @next */
void send_with_flags(uint8_t next, uint8_t src, uint8_t dst, uint8_t flags,
                     uint8_t *data, size_t len, uint8_t ttl)
{
    frame_t *f;

    beginSend();

    f = (frame_t*)TRX_BUFFER_ADDR(3);
    TRX_FRAME_BUFFER(0) = sizeof(*f) + 2 + 2 + len;
    TRX_FRAME_BUFFER(1) = node_address;
    TRX_FRAME_BUFFER(2) = next;

    f->src = src;
    f->dst = dst;
    f->flags = flags;
    f->ttl = ttl;
    if (len)
        memcpy(f->data, data, len);

    finishSend();
}

void  broadcast_time(uint16_t next_sleep)
{
    frame_t *f;
    uint8_t *data;
    uint32_t scnt;

    beginSend();

    f = (frame_t*)TRX_BUFFER_ADDR(3);
    /* frame size + 2 for CRC + 2 for link level address */
    TRX_FRAME_BUFFER(0) = sizeof(*f) + 2 + 2 + sizeof(next_sleep) + sizeof(scnt);
    TRX_FRAME_BUFFER(1) = node_address;
    TRX_FRAME_BUFFER(2) = 0xFF;

    f->src = node_address;
    f->dst = 0xFF;
    f->flags = FLAG_TIME;

    scnt = get_time_ms();

    /* This is quite hacky. Send a uint16_t and a uint32_t */
    data = (uint8_t*)f->data;
    memcpy(data, &next_sleep, sizeof(next_sleep));
    data += sizeof(next_sleep);
    memcpy(data, &scnt, sizeof(scnt));

    finishSend();
}

/* Listen to the channel, before sending a frame to the recipient */
int isChannelIdle(uint8_t rcpt)
{
	int i;

	/* Wait for about 50 msec */
	for (i = 0; i < 5; i++) {
		if (rxLen && (rxFrame->flags & FLAG_RTS) == FLAG_RTS && rxFrame->src == rcpt)
			return 0;
		_delay_ms(10);
	}
	return 1; //IDLE
}

/* Wait a random time, before a retry */
void random_backoff()
{
	int time = rand() % 100;
    if (time < 25)
    	_delay_ms(25);
    else if (time < 50)
        _delay_ms(50);
    else if (time < 75)
        _delay_ms(75);
    else
        _delay_ms(100);
}

void send_packet(uint8_t src, uint8_t dst, uint8_t *data, size_t sz, uint8_t ttl)
{
    int i;
    uint8_t next;

	DevelDebug("[%s] Started\n", __FUNCTION__);

    /* Find next hop for @dst */
    next = find_route(dst);
    if (next == 0xFF) {
        /* TODO: handle no route */
        return;
    }
	/* Colision detection
     * Listen for another RTS packet. Backoff if one is received
	 */
	if (isChannelIdle(next)) {
		/* Send RTS */
		DevelDebug("[%s] Send RTS to %d through %d\n", __FUNCTION__, dst, next);
		DevelDebug("[%s] rxFrame->src=%d rxFrame->dst=%d rxFrame->flags=%d\n",
                   __FUNCTION__, rxFrame->src, rxFrame->dst, rxFrame->flags);
		send_with_flags(next, src, dst, FLAG_RTS, NULL, 0, 1);
	} else {
		/* Wait for a random backoff time */
		DevelDebug("[%s] Backoff! Channel busy\n", __FUNCTION__);
		random_backoff();
	}
	/* Wait for CTS */
	DevelDebug("[%s] Wait for CTS\n", __FUNCTION__);
    setState(CMD_RX_ON);
    for (i = 0; i< 10; i++) {
        if (rxLen) {
            /* Send actual data */
            if (rxFrame->src == src && rxFrame->dst == dst && (rxFrame->flags & FLAG_CTS)) {
				DevelDebug("[%s] Got CTS\n", __FUNCTION__);
                _delay_ms(20);
				DevelDebug("[%s] Send DATA to %d\n", __FUNCTION__, next);
                send_with_flags(next, src, dst, FLAG_NONE, data, sz, ttl);
                break;
            }
        }
        _delay_ms(30);
    }
    DevelDebug("[%s] Ended\n", __FUNCTION__);
}

uint8_t find_route(uint8_t dst)
{
    int i;
    int besti;
    uint8_t minhop;

    besti = -1;
    minhop = TTL_MAX;
    for (i = 0; i < MAX_ROUTES; i++) {
        if (!rtable[i].used)
            continue;
        if (rtable[i].dst == dst && rtable[i].hops < minhop) {
            besti = i;
            minhop = rtable[i].hops;
        }
    }
    if (besti >= 0)
            return rtable[besti].next;

    return 0xFF;
}

uint8_t add_route(uint8_t dst, uint8_t next, uint8_t ttl)
{
    int i;

    for (i = 0; i < MAX_ROUTES; i++) {
        if (rtable[i].used)
            continue;
        break;
    }
    if (i == MAX_ROUTES)
        return 0xFF;
    rtable[i].dst = dst;
    rtable[i].next = next;
    rtable[i].hops = TTL_MAX - ttl;
    rtable[i].used = 1;

    return i;
}

uint8_t updateRTable(uint8_t sender, volatile frame_t *f)
{
    int i;

    /* Add connected route */
    i = find_route(sender);
    if (i == 0xFF) {
        add_route(sender, sender, f->ttl);
    }
    /* Add remote route */
    i = find_route(f->src);
    if (i == 0xFF)
        i = add_route(f->src, sender, f->ttl);

    return i;
}


void update_time()
{
    uint16_t rsleep;
    uint32_t rcnt;

    memcpy(&rsleep, (const void*)rxFrame->data, sizeof(rsleep));
    memcpy(&rcnt, (const void*)(rxFrame->data + sizeof(rsleep)), sizeof(rcnt));

    int i = 0;
    for (i = 0; i < MAX_ROUTES; i++)
	if (rtable[i].used && rtable[i].dst == rxSender)
		break;

    rtable[i].dst_time = rcnt;
    rtable[i].dst_sleep = rsleep;
    rtable[i].this_time = rxTime;
    rtable[i].used = 1;
    //printf("TIME[%d]@%lu: cnt: %lu sleep: %u\n", rxSender, rtable[i].this_time, rcnt, rsleep);
}

void process_packet(uint8_t src, uint8_t dst)
{
    size_t frame_size = rxLen - 2;
    frame_t *copy = malloc(frame_size);

    memcpy(copy, (const void*)rxFrame, frame_size);

	DevelDebug("[%s] Process packet sent by %d to %d\n", __FUNCTION__, src, dst);
	DevelDebug("[%s] rxFrame->src=%d rxFrame->dst=%d rxFrame->flags=%d\n", __FUNCTION__, rxFrame->src, rxFrame->dst, rxFrame->flags);
    /* Am I the destination of the packet? */
    if (rxFrame->dst == node_address) {
		DevelDebug("[%s] Packet reached DESTINATION\n", __FUNCTION__);
        if ((rxFrame->flags & FLAG_RTS) == FLAG_RTS) {
			/* RTS received, so send CTS */
			DevelDebug("[%s] Send CTS to %d\n", __FUNCTION__, rxSender);
            send_with_flags(rxSender, rxFrame->src, rxFrame->dst, FLAG_CTS, NULL, 0, 1);
        } else {
			/* Data received */
            if (rxFrame->flags == FLAG_NONE) {
                /* TODO: Do something with it*/
                DevelDebug("[%s] Consume data\n", __FUNCTION__);
                printStringLn("consume");
            }
        }
    }
	/* Or I am just a hop */
	else {
        if (rxRecipient == node_address) {
			DevelDebug("[%s] Packet reached a HOP\n", __FUNCTION__);
            if ((rxFrame->flags & FLAG_RTS) == FLAG_RTS) {
				/* RTS received, so send CTS */
				DevelDebug("[%s] Send CTS to %d\n", __FUNCTION__, rxSender);
                send_with_flags(rxSender, rxFrame->src, rxFrame->dst, FLAG_CTS, NULL, 0, 1);
            } else {
                /* Forward packet */
                printf("COPY TTL: %u\n", copy->ttl);
                DevelDebug("[%s] Forward packet:%x sent by %d to %d\n", __FUNCTION__, frame_size, copy->src, copy->dst);
                send_packet(copy->src, copy->dst, (uint8_t*)copy->data, frame_size - sizeof(*copy) - 2,
                            copy->ttl - 1);
            }
        } else {
            // Time Broadcast
            if ((rxFrame->flags & FLAG_TIME) == FLAG_TIME) {
                update_time();
                /* TIME packet received */
                printStringLn("TIME Received");
            }
        }
    }
    free(copy);
}

uint32_t calc_next_sleep_time()
{
    uint32_t new_sleep_time;
    uint32_t delay, after;
    uint8_t count;
    int i;

    delay = 0;

    new_sleep_time = next_sleep;
    count = 1;
    for (i = 0; i < MAX_ROUTES; i++) {
        /* Only existing entries */
        if (!rtable[i].used)
            continue;
        /* Only direct neighbours */
        if (rtable[i].dst != rtable[i].next)
            continue;

        // calculate next wake-up time
        //new_sleep_time += rtable[i].dst_sleep + rtable[i].dst_time - rtable[i].this_time;

        // calculate average delay between neighbours
        delay += rtable[i].this_time - rtable[i].dst_time;

        count++;
    }

    after = get_time_ms();
    delay /= (count - 1);

#ifndef WAKEUP_SYNC
    set_symbol_counter(after + delay);
#endif

    for (i = 0; i < MAX_ROUTES; i++) {
        /* Only existing entries */
        if (!rtable[i].used)
            continue;
        /* Only direct neighbours */
        if (rtable[i].dst != rtable[i].next)
            continue;

        // calculate next wake-up time
        new_sleep_time += rtable[i].dst_sleep + rtable[i].dst_time - rtable[i].this_time;
    }
    new_sleep_time /= count;
    new_sleep_time += delay;

    printf("Adjusted local time: %lu -> %lu\n", after, after + delay);

    return MAX(MIN_SLEEP_MS, new_sleep_time);
}

void print_rtable()
{
    int i;

    printf("%8s\t%8s\t%8s\t%8s\t%8s\t%8s\t%8s\n", "dst", "next", "used",
           "this_time", "dst_time", "dst_sleep", "hops");
    for (i = 0; i < MAX_ROUTES; i++) {
        if (rtable[i].used == 0)
            continue;

        printf("%8d\t%8d\t%8d\t", rtable[i].dst, rtable[i].next, rtable[i].used);
        printf("%8lu\t%8lu\t%8u\t", rtable[i].this_time, rtable[i].dst_time, rtable[i].dst_sleep);
	    printf("%8u\n", rtable[i].hops);
    }
}

/* Assemble a frame based on received parameters */
void assemble_frame(frame_t* f_assembled, uint8_t src, uint8_t dst, uint8_t* data, size_t len, uint8_t flags)
{
	f_assembled->src = src;
	f_assembled->dst = dst;
	if (len)
		memcpy(f_assembled->data, data, len);
	f_assembled->flags = flags;
}

/* Check if the frame is a CTS one and is received as a reply of RTS */
int isExpectedCTS(frame_t* f_recv, uint8_t src, uint8_t dst)
{
	if ((f_recv->flags == FLAG_CTS)
		&&(f_recv->src == src)
		&&(f_recv->dst == dst)
		) {
		return 1;
	}
	return 0;
}

int main()
{
    int value;

    initLeds();
    initUart();
    rfInit();

	DDRE = 0xff;
    PORTE = 0xff;
	/* ADC power down */
	//PRR0 &= ~_BV(PRADC);
	//DDRB &= ~(_BV(PD0) | _BV(PD1));
	//DDRB |= _BV(PD0);
    initializeSymbolCounter();

    rxLen = 0;
    value = 0;

    printf("Started: %lu\n", get_symbol_counter());
    setState(CMD_RX_ON);

    next_sleep = 7 * 1000 + rand() % 1000;

	/* BROADCAST Stage */
	printf("Broadcast started\n");
    int j = 4;
    while(j >= 0) {
        printf("Wake-up at %lu\n", get_symbol_counter());
        delaySymbolCounterMS(next_sleep);
        PORTB |= _BV(PB5);

        if (node_address == j) {
            _delay_ms(500);
            sendBroadcast(j, 0xFF, TTL_MAX);
            PORTB ^= (1<<PB7);
            printf("Broadcast Sent at %lu\n", get_symbol_counter());
            PORTB ^= (1<<PB7);
            _delay_ms(1480);
        } else {
			/* Delay, until receiving a packet */
            int i;
            for (i = 0; i < 199; i++) {
                if (rxLen) {
                    printStringLn("Broadcast received");

                    /* Update routing table */
                    updateRTable(rxSender, rxFrame);
                    /* Forward the broadcast */
                    sendBroadcast(rxFrame->src, rxFrame->dst, rxFrame->ttl - 1);
                    print_rtable();
                    rxLen = 0;
                }
                _delay_ms(10);
            }
            setState(CMD_RX_ON);
        }
        j--;
        //_delay_ms(5000);
        printStringLn("Enter in sleep");
        //_delay_ms(5000);
        PORTB &= ~_BV(PB5);

        /* Sleep transceiver */
        TRX_STATE = CMD_FORCE_TRX_OFF;
        TRXPR = _BV(SLPTR);

        set_sleep_mode(SLEEP_MODE_PWR_SAVE);
        sleep_enable();
		/* TODO: This sleep_cpu() is problematic for NOW */
        sleep_cpu();
        sleep_disable();

        /* Exit from sleep transceiver */
        TRXPR &= ~_BV(SLPTR);
        TRX_STATE = CMD_TRX_OFF;

        value++;
        setState(CMD_FORCE_TRX_OFF);
        setState(CMD_RX_ON);
    }
	//TRX_STATE = CMD_FORCE_TRX_OFF;
    setState(CMD_FORCE_TRX_OFF);
    setState(CMD_RX_ON);

resync:
    printStringLn("TIME SYNC START");

    /* Time Sync */
    value = 4;
    printf("Initial Sleep: %lu\n", next_sleep);
    while (value >= 0) {
        if (node_address == value) {
            _delay_ms(500);
            //uint32_t before = get_time_ms();
            broadcast_time(next_sleep);
            //uint32_t after = get_time_ms();
            // better solution
            //_delay_ms(1000 - (after - before));
            _delay_ms(1490);
        } else {
            printf("Wait for packet in state %d\n", TRX_STATUS & 0x1F);
			/* Delay, until receiving a packet */
            int i;
            for (i = 0; i < 190; i++) {
                if (rxLen) {
                    /* Update routing table */
                    updateRTable(rxSender, rxFrame);

                    process_packet(rxSender, rxRecipient);
                    print_rtable();
                    printStringLn("Message received");
                    rxLen = 0;
                    setState(CMD_FORCE_TRX_OFF);
                    setState(CMD_RX_ON);
                }
                _delay_ms(10);
            }
            setState(CMD_FORCE_TRX_OFF);
            setState(CMD_RX_ON);
        }

        value--;
        printStringLn("Enter in sleep");
        delaySymbolCounterMS(next_sleep);

        /* Sleep transceiver */
        TRX_STATE = CMD_FORCE_TRX_OFF;
        TRXPR = _BV(SLPTR);

        set_sleep_mode(SLEEP_MODE_PWR_SAVE);
        sleep_enable();
		/* TODO: This sleep_cpu() is problematic for NOW */
        sleep_cpu();
        sleep_disable();

        /* Exit from sleep transceiver */
        TRXPR &= ~_BV(SLPTR);
        TRX_STATE = CMD_TRX_OFF;

        setState(CMD_FORCE_TRX_OFF);
        setState(CMD_RX_ON);
    }

    printStringLn("TIME SYNC END");

#ifdef WAKEUP_SYNC
    next_sleep = calc_next_sleep_time();
#else
    calc_next_sleep_time();
#endif

    printf("Next Sleep: %lu\n", next_sleep);

    //goto resync;

	/* COMMUNICATION Stage */
	printf("Communication started\n");
    j = 1;
    while (j--) {
        //delaySymbolCounter(7);
        if (node_address == 2) {
            _delay_ms(500);
            uint8_t msg[] = {0xaa, 0xbb, 0xcc, 0xdd};
            send_packet(node_address, 1, msg, sizeof(msg), TTL_MAX);
            _delay_ms(1480);
        } else {
            printf("Wait for packet in state %d\n", TRX_STATUS & 0x1F);
			/* Delay, until receiving a packet */
            int i;
            for (i = 0; i < 190; i++) {
                if (rxLen) {
                    /* Update routing table */
                    updateRTable(rxSender, rxFrame);

                    process_packet(rxSender, rxRecipient);
                    printStringLn("Message received");
                    rxLen = 0;
                    setState(CMD_FORCE_TRX_OFF);
                    setState(CMD_RX_ON);
                }
                _delay_ms(10);
            }
            setState(CMD_FORCE_TRX_OFF);
            setState(CMD_RX_ON);
        }

        printStringLn("Enter in sleep");
        //_delay_ms(5000);
        //delaySymbolCounter(7);
        delaySymbolCounterMS(next_sleep);

        /* Sleep transceiver */
        TRX_STATE = CMD_FORCE_TRX_OFF;
        TRXPR = _BV(SLPTR);

        set_sleep_mode(SLEEP_MODE_PWR_SAVE);
        sleep_enable();
		/* TODO: This sleep_cpu() is problematic for NOW */
        sleep_cpu();
        sleep_disable();

        /* Exit from sleep transceiver */
        TRXPR &= ~_BV(SLPTR);
        TRX_STATE = CMD_TRX_OFF;

        setState(CMD_FORCE_TRX_OFF);
        setState(CMD_RX_ON);
    }

    goto resync;

    return 0;
}
